from html.parser import HTMLParser
from ..tags import Element

class PYXParser(HTMLParser):
	def __init__(self):
		self.document = Element("html")
		self.current_node: Element = self.document
		self.current_tag = ""
		super().__init__(convert_charrefs=True)

	def handle_starttag(self, tag: str, attrs: list[str]):
		dictattrs = {}
		for attr in attrs:
			if attr[1] is None:
				dictattrs[attr[0]] = '""'
				continue

			if attr[1].startswith('{') and attr[1].endswith('}'):
				dictattrs[attr[0]] = attr[1].removeprefix('{').removesuffix('}')
				continue
		
			dictattrs[attr[0]] = repr(attr[1])


		self.current_tag = tag
		element = Element(tag, data="", children=None, **dictattrs)
		
		self.current_node.children.append(element)
		element.parent = self.current_node

		self.current_node = element

	def handle_endtag(self, tag: str):	
		if tag != self.current_tag:
			raise ValueError(f"The tag '{self.current_tag}' was never closed!")

		self.current_node = self.current_node.parent
		# self.current_node is already set to parent
		self.current_tag = self.current_node.tag

	def parse(self, feed: str):
		super().feed(feed)

		if (self.current_node is not self.document):
			raise ValueError("A tag wasn't closed!")

	def handle_data(self, data):
		self.current_node.data = 'f'+repr(data) # act like it is a fmt string



from sys import argv
from typing import Literal, Union
from ..tags import Element
from .generator import PyCodeGenerator
from .xmlparser import PYXParser

def main():
	file: str = argv[1]

	with open(file, 'r') as f:
		code = f.read()

	new_code = ""
	in_double_quote_str = False
	in_single_quote_str = False
	in_esc_code = False
	in_pyx = False
	current_pyx_start: int = None
	pyx_idxs: list[tuple[int, int]] = []
	pyx_blocks: list[str] = []

	def check_start_pyx(code: str, i: int) -> bool:
		if code[i:i+5] == "<pyx>": return True
		return False

	def check_end_pyx(code: str, i: int) -> bool:
		if code[i:i+6] == "</pyx>": return True
		return False

	def strchecker(char: Union[Literal["'"], Literal['"']]) -> int:
		if char == '"':
			if in_double_quote_str and in_esc_code:
				return -1
			if not in_double_quote_str:
				return 1
			
			return 0

		if char == "'":
			if in_single_quote_str and in_esc_code:
				return -1
			if not in_single_quote_str:
				return 1
			
			return 0

	obj = enumerate(code)
	for i, char in obj:
		if in_esc_code:
			in_esc_code = False
			continue

		if (char in ('"', "'")) and (not in_pyx):
			val = strchecker(char)
			if (val == -1): continue

			if in_esc_code:
				in_esc_code = False

			elif char == '"':
				in_double_quote_str = val

			elif char == "'":
				in_single_quote_str = val

		if (char == '\\') and (in_double_quote_str or in_single_quote_str):
			if in_esc_code:
				if not in_pyx: continue
			else:
				in_esc_code = 1

		if char == '<' and not (in_double_quote_str or in_single_quote_str):
			if not in_pyx:
				res = check_start_pyx(code, i)

				if res:
					current_pyx_start = i
					for i in range(5): next(obj)

					in_pyx = True
					pyx_blocks.append("")
				
				continue

			res = check_end_pyx(code, i)

			if res:
				pyx_idxs.append(
					(current_pyx_start, i+7)
				)

				for i in range(7): next(obj)

				in_pyx = False

		if in_pyx:
			pyx_blocks[-1]+=char
			continue

	documents: list[Element] = []

	for block in pyx_blocks:
		parser = PYXParser()
		parser.parse(block)
		documents.append(
			parser.document
		)

	pycodes = [PyCodeGenerator(document).generate() for document in documents]

	prevend = 0

	for i, pycode in enumerate(pycodes):
		start = pyx_idxs[i][0]
		end = pyx_idxs[i][1]

		new_code+=code[prevend:start]
		new_code+=pycode

		prevend = end

		if i == len(pycodes)-1:
			new_code+=code[end:]

	if not new_code: new_code = code

	with open('.'.join(argv[1].split('.')[:-1])+".py", 'w') as f:
		f.write(new_code)

if __name__ == "__main__":
	main()


from ..tags import Element

class PyCodeGenerator():
	def __init__(self, document: Element):
		self.document = document

	def generate(self, element: Element=None) -> str:
		element = element if element else self.document

		children = [self.generate(elem) for elem in element.children]
		attrs = (
			'{'+(
				','.join(
					f"{repr(name)}: {value}" 
					for name, value in element.attrs.items()
				)
			)+'}'
		)

		data = f"data={element.data}"

		return \
			f"{element.tag}({data+',' if element.data else ''} " \
				f"children=[{','.join(children)}], **{attrs})"


standard_tags = [
	"a",
	"abbr",
	"acronym",
	"address",
	"applet",
	"area",
	"article",
	"aside",
	"audio",
	"b",
	"base",
	"basefont",
	"bdi",
	"bdo",
	"bgsound",
	"big",
	"blink",
	"blockquote",
	"body",
	"br",
	"button",
	"canvas",
	"caption",
	"center",
	"cite",
	"code",
	"col",
	"colgroup",
	"content",
	"data",
	"datalist",
	"dd",
	"decorator",
	"del",
	"details",
	"dfn",
	"dir",
	"div",
	"dl",
	"dt",
	"element",
	"em",
	"embed",
	"fieldset",
	"figcaption",
	"figure",
	"font",
	"footer",
	"form",
	"frame",
	"frameset",
	"h1",
	"h2",
	"h3",
	"h4",
	"h5",
	"h6",
	"head",
	"header",
	"hgroup",
	"hr",
	"html",
	"i",
	"iframe",
	"img",
	"input",
	"ins",
	"isindex",
	"kbd",
	"keygen",
	"label",
	"legend",
	"li",
	"link",
	"listing",
	"main",
	"map",
	"mark",
	"marquee",
	"menu",
	"menuitem",
	"meta",
	"meter",
	"nav",
	"nobr",
	"noframes",
	"noscript",
	"object",
	"ol",
	"optgroup",
	"option",
	"output",
	"p",
	"param",
	"plaintext",
	"pre",
	"progress",
	"q",
	"rp",
	"rt",
	"ruby",
	"s",
	"samp",
	"script",
	"section",
	"select",
	"shadow",
	"small",
	"source",
	"spacer",
	"span",
	"strike",
	"strong",
	"style",
	"sub",
	"summary",
	"sup",
	"table",
	"tbody",
	"td",
	"template",
	"textarea",
	"tfoot",
	"th",
	"thead",
	"time",
	"title",
	"tr",
	"track",
	"tt",
	"u",
	"ul",
	"var",
	"video",
	"wbr",
	"xmp"
]

class Element:
	def __init__(self, tag: str, data: str="", children: list=None, **attrs: str):
		self.tag = tag
		self.data = data
		self.children: list[Element] = children if children else []
		self.attrs = attrs
		self.parent: Element = None

	@property
	def html(self) -> str:
		return str(self)

	def __str__(self):
		attrs = ""
		children = '\n'.join(str(child) for child in self.children)

		for attrname, value in self.attrs.items():
			attrs+=f'{attrname}="{value}" '

		return f"<{self.tag} {attrs}>{self.data}{children}</{self.tag}>"

for tag in standard_tags:
	globals()[tag] = type(
		tag, 
		(Element,),
		{
			"__init__": 
			lambda self, data="", children=None, **attrs:
			super(self.__class__, self).__init__(self.__class__.__name__, data, children, **attrs)
		}
	)


import sys
import os

def main():
	if len(sys.argv) < 2: exit(1)

	compiled_files = []

	if sys.argv[1] == "run":
		for path, dirnames, filenames in os.walk(os.getcwd()):
			filename: str
			for filename in filenames:
				if filename.endswith(".pyx"):
					compiled_files.append(
						'.'.join(
							os.path.join(path, filename)
								.split('.')[-1]
						)+".py"
					)
					os.system(f"pyxc {os.path.join(path, filename)}")

		try: os.system("flask run")
		except KeyboardInterrupt: pass

	if sys.argv[1] == "new":
		if len(sys.argv) < 3:
			print("Missing project name!")
			exit(1)

		proj_name = sys.argv[2]

		try:
			os.mkdir(proj_name)
			os.mkdir(f"{proj_name}/components")
			os.mkdir(f"{proj_name}/snippets")

			with open(f"{proj_name}/app.pyx", 'w') as f:
				f.write(
					'import pysite\nfrom flask import Flask\nfrom pysite.tags '
					'import *\n\napp = Flask(__name__)\n\n@app.route(\'/\')\n'
					'def index():\n\treturn (\n\t\t<pyx>\n\t\t\t<h1>'
					'Hello World!</h1>\n\t\t\t<br></br>\n\t\t\t<p '
					'style="display:inline-block;">This site was built with</p>'
					'\n\t\t\t<a href="https://github.com/User0332/pyx">PySite'
					'</a>\n\t\t</pyx>\n\t).html'
				)

			open(f"{proj_name}/components/__init__.py", 'w').close()
			open(f"{proj_name}/snippets/__init__.py", 'w').close()
		except OSError as e:
			print(f"An error occurred: {e}")

if __name__ == "__main__":
	main()